@@ -1,7 +1,6 @@
 #include <sys/types.h>
 
 #include <ctype.h>
-#include <err.h>
 #include <errno.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -58,7 +57,8 @@ enum TagId {
 	RSSTagAuthor, RSSTagDccreator,
 	RSSTagCategory,
 	/* Atom */
-	AtomTagUpdated, AtomTagPublished, /* creation date has higher priority */
+	/* creation date has higher priority */
+	AtomTagModified, AtomTagUpdated, AtomTagIssued, AtomTagPublished,
 	AtomTagTitle,
 	AtomTagMediaDescription, AtomTagSummary, AtomTagContent,
 	AtomTagId,
@@ -151,9 +151,11 @@ static FeedTag atomtags[] = {
 	{ STRP("category"),          AtomTagCategory         },
 	{ STRP("content"),           AtomTagContent          },
 	{ STRP("id"),                AtomTagId               },
+	{ STRP("issued"),            AtomTagIssued           }, /* Atom 0.3 */
 	/* Atom: <link href="" />, RSS has <link></link> */
 	{ STRP("link"),              AtomTagLink             },
 	{ STRP("media:description"), AtomTagMediaDescription },
+	{ STRP("modified"),          AtomTagModified         }, /* Atom 0.3 */
 	{ STRP("published"),         AtomTagPublished        },
 	{ STRP("summary"),           AtomTagSummary          },
 	{ STRP("title"),             AtomTagTitle            },
@@ -186,7 +188,9 @@ static int fieldmap[TagLast] = {
 	[RSSTagDccreator]          = FeedFieldAuthor,
 	[RSSTagCategory]           = FeedFieldCategory,
 	/* Atom */
+	[AtomTagModified]          = FeedFieldTime,
 	[AtomTagUpdated]           = FeedFieldTime,
+	[AtomTagIssued]            = FeedFieldTime,
 	[AtomTagPublished]         = FeedFieldTime,
 	[AtomTagTitle]             = FeedFieldTitle,
 	[AtomTagMediaDescription]  = FeedFieldContent,
@@ -209,8 +213,7 @@ static const char *baseurl;
 
 static FeedContext ctx;
 static XMLParser parser; /* XML parser state */
-static String tmpstr;
-static enum ContentType tmpcontenttype; /* content-type for item */
+static String attrispermalink, attrrel, attrtype, tmpstr;
 
 int
 tagcmp(const void *v1, const void *v2)
@@ -702,18 +705,8 @@ xmlattr(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl,
 	/* content-type may be: Atom: text, xhtml, html or mime-type.
 	   MRSS (media:description): plain, html. */
 	if (ISCONTENTTAG(ctx)) {
-		if (isattr(n, nl, STRP("type"))) {
-			if (isattr(v, vl, STRP("html")) ||
-			    isattr(v, vl, STRP("xhtml")) ||
-			    isattr(v, vl, STRP("text/html")) ||
-			    isattr(v, vl, STRP("text/xhtml"))) {
-				tmpcontenttype = ContentTypeHTML;
-			} else if (isattr(v, vl, STRP("text")) ||
-			           isattr(v, vl, STRP("plain")) ||
-				   isattr(v, vl, STRP("text/plain"))) {
-				tmpcontenttype = ContentTypePlain;
-			}
-		}
+		if (isattr(n, nl, STRP("type")))
+			string_append(&attrtype, v, vl);
 		return;
 	}
 
@@ -721,30 +714,15 @@ xmlattr(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl,
 		if (ctx.tag.id == RSSTagEnclosure &&
 		    isattr(n, nl, STRP("url"))) {
 			string_append(&tmpstr, v, vl);
-		} else if ((ctx.tag.id == RSSTagGuid ||
-		            ctx.tag.id == RSSTagGuidPermalinkFalse ||
-			    ctx.tag.id == RSSTagGuidPermalinkTrue) &&
+		} else if (ctx.tag.id == RSSTagGuid &&
 		           isattr(n, nl, STRP("ispermalink"))) {
-			if (isattr(v, vl, STRP("true")))
-				ctx.tag.id = RSSTagGuidPermalinkTrue;
-			else
-				ctx.tag.id = RSSTagGuidPermalinkFalse;
+			string_append(&attrispermalink, v, vl);
 		}
 	} else if (ctx.feedtype == FeedTypeAtom) {
-		if (ctx.tag.id == AtomTagLink ||
-		    ctx.tag.id == AtomTagLinkAlternate ||
-		    ctx.tag.id == AtomTagLinkEnclosure) {
+		if (ctx.tag.id == AtomTagLink) {
 			if (isattr(n, nl, STRP("rel"))) {
-				/* empty or "alternate": other types could be
-				   "enclosure", "related", "self" or "via" */
-				if (!vl || isattr(v, vl, STRP("alternate")))
-					ctx.tag.id = AtomTagLinkAlternate;
-				else if (isattr(v, vl, STRP("enclosure")))
-					ctx.tag.id = AtomTagLinkEnclosure;
-				else
-					ctx.tag.id = AtomTagLink; /* unknown */
-			} else if (ctx.tag.id != AtomTagLink &&
-			           isattr(n, nl, STRP("href"))) {
+				string_append(&attrrel, v, vl);
+			} else if (isattr(n, nl, STRP("href"))) {
 				string_append(&tmpstr, v, vl);
 			}
 		} else if (ctx.tag.id == AtomTagCategory &&
@@ -806,6 +784,18 @@ xmlattrstart(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl)
 		}
 		return;
 	}
+
+	if (attrispermalink.len && isattr(n, nl, STRP("ispermalink")))
+		string_clear(&attrispermalink);
+	else if (attrrel.len && isattr(n, nl, STRP("rel")))
+		string_clear(&attrrel);
+	else if (attrtype.len && isattr(n, nl, STRP("type")))
+		string_clear(&attrtype);
+	else if (tmpstr.len &&
+	    (isattr(n, nl, STRP("href")) ||
+	     isattr(n, nl, STRP("term")) ||
+	     isattr(n, nl, STRP("url"))))
+		string_clear(&tmpstr); /* use the last value for multiple attribute values */
 }
 
 /* NOTE: this handler can be called multiple times if the data in this
@@ -880,29 +870,10 @@ xmltagstart(XMLParser *p, const char *t, size_t tl)
 		memcpy(&(ctx.tag), f, sizeof(ctx.tag));
 	}
 
-	switch (ctx.tag.id) {
-	case AtomTagLink:
-		/* without a rel attribute the default link type is "alternate" */
-		ctx.tag.id = AtomTagLinkAlternate;
-		break;
-	case RSSTagGuid:
-		/* without a ispermalink attribute the default value is "true" */
-		ctx.tag.id = RSSTagGuidPermalinkTrue;
-		break;
-	case RSSTagContentEncoded:
-	case RSSTagDescription:
-		tmpcontenttype = ContentTypeHTML; /* default content-type */
-		break;
-	case RSSTagMediaDescription:
-	case AtomTagContent:
-	case AtomTagMediaDescription:
-	case AtomTagSummary:
-		tmpcontenttype = ContentTypePlain; /* default content-type */
-		break;
-	default:
-		break;
-	}
 	ctx.iscontenttag = (fieldmap[ctx.tag.id] == FeedFieldContent);
+	string_clear(&attrispermalink);
+	string_clear(&attrrel);
+	string_clear(&attrtype);
 }
 
 static void
@@ -920,6 +891,25 @@ xmltagstartparsed(XMLParser *p, const char *t, size_t tl, int isshort)
 		return;
 	}
 
+	/* set tag type based on it's attribute value */
+	if (ctx.tag.id == RSSTagGuid) {
+		/* if empty the default is "true" */
+		if (!attrispermalink.len ||
+		    isattr(attrispermalink.data, attrispermalink.len, STRP("true")))
+			ctx.tag.id = RSSTagGuidPermalinkTrue;
+		else
+			ctx.tag.id = RSSTagGuidPermalinkFalse;
+	} else if (ctx.tag.id == AtomTagLink) {
+		/* empty or "alternate": other types could be
+		   "enclosure", "related", "self" or "via" */
+		if (!attrrel.len || isattr(attrrel.data, attrrel.len, STRP("alternate")))
+			ctx.tag.id = AtomTagLinkAlternate;
+		else if (isattr(attrrel.data, attrrel.len, STRP("enclosure")))
+			ctx.tag.id = AtomTagLinkEnclosure;
+		else
+			ctx.tag.id = AtomTagLink; /* unknown */
+	}
+
 	tagid = ctx.tag.id;
 
 	/* map tag type to field: unknown or lesser priority is ignored,
@@ -933,7 +923,24 @@ xmltagstartparsed(XMLParser *p, const char *t, size_t tl, int isshort)
 	if (ctx.iscontenttag) {
 		ctx.iscontent = 1;
 		ctx.iscontenttag = 0;
-		ctx.contenttype = tmpcontenttype;
+
+		/* detect content-type based on type attribute */
+		if (attrtype.len) {
+			if (isattr(attrtype.data, attrtype.len, STRP("html")) ||
+			    isattr(attrtype.data, attrtype.len, STRP("xhtml")) ||
+			    isattr(attrtype.data, attrtype.len, STRP("text/html")) ||
+			    isattr(attrtype.data, attrtype.len, STRP("text/xhtml")) ||
+			    isattr(attrtype.data, attrtype.len, STRP("application/xhtml+xml")))
+				ctx.contenttype = ContentTypeHTML;
+			else /* unknown: handle as base64 text data */
+				ctx.contenttype = ContentTypePlain;
+		} else {
+			/* default content-type */
+			if (tagid == RSSTagContentEncoded || tagid == RSSTagDescription)
+				ctx.contenttype = ContentTypeHTML;
+			else
+				ctx.contenttype = ContentTypePlain;
+		}
 	}
 
 	ctx.field = &(ctx.fields[fieldmap[tagid]].str);
@@ -966,7 +973,7 @@ xmltagend(XMLParser *p, const char *t, size_t tl, int isshort)
 	} else if (ctx.tag.id && istag(ctx.tag.name, ctx.tag.len, t, tl)) {
 		/* matched tag end: close it */
 		/* copy also to the link field if the attribute isPermaLink="true"
-		    and it is not set by a tag with higher prio. */
+		   and it is not set by a tag with higher prio. */
 		if (ctx.tag.id == RSSTagGuidPermalinkTrue && ctx.field &&
 		    ctx.tag.id > ctx.fields[FeedFieldLink].tagid) {
 			string_clear(&ctx.fields[FeedFieldLink].str);
